/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvworker.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.
 
#include <iostream>

#include <QDebug>
#include <QMutex>
#include <QWaitCondition>
#include <QApplication>

#include <QVWorker>

QVWorker::QVWorker(const QString name):QVPropertyContainer(name), numIterations(0), status(Running), triggerList(), minms(0)
	{
	qDebug() << "QVWorker::QVWorker(" << name << ")";
	Q_ASSERT_X(qvApp != NULL, "QVWorker::QVWorker()", "QVApplication doesn't exists");
	if (qvApp == NULL)
		{
		QString str = "QVWorker::QVWorker(): the QVWorker cannot be created before the QVApplication instance. Aborting now.";
		std::cerr << qPrintable(str) << std::endl;
		exit(1);
		}

	addProperty<int>("max worker iterations", inputFlag | guiInvisible | internalProp, -1, "Stablishes maximal number of iterations to execute worker");
	maxIterations = getPropertyValue<int>("max worker iterations");

	addProperty<bool>("stats enabled", inputFlag | guiInvisible | internalProp, TRUE, "Stablishes if the worker's cpu stats will be enabled");
	statsEnabled = getPropertyValue<bool>("stats enabled");
	if (statsEnabled) cpuStatControler = new QVStatControler();
	if (statsEnabled)	addProperty<QVStat>("cpu stats", outputFlag, cpuStatControler->value(), "CPU stats's Statistics");
	else				addProperty<QVStat>("cpu stats", outputFlag, QVStat(), "CPU stats's Statistics");


// 	informer->setParent(this); // set the informer's parent = this, in order to it will be moved to the worker's thread when the qvApp do worker->moveToThread(worker)
	qDebug() << "QVWorker::QVWorker(" << name << ") <- return";
	};

QVWorker::QVWorker(const QVWorker &other):QThread(), QVPropertyContainer(other), statsEnabled(other.statsEnabled), numIterations(other.numIterations),
	maxIterations(other.maxIterations), status(other.status), triggerList(other.triggerList), iterationTime(other.iterationTime), curms(other.curms),
	minms(other.minms)
	{
	if (statsEnabled) cpuStatControler = new QVStatControler();
	}

QVWorker::~QVWorker()
	{
	if (statsEnabled)
		delete cpuStatControler;

// 	if (informer != 0) delete informer;
	}
/*
void QVWorker::unlink()
	{
	if(status == Finished)
		QVPropertyContainer::unlink();
	else
		std::cerr << "WARNING: A worker only can be unlinked if the worker's status is Finished." << std::endl;
	}
*/
void QVWorker::run()
	{
	qDebug() << "QVWorker::run()";

	while(status != Finished)
		{
		qDebug() << "Processing events in worker " << qPrintable(getName());

		// First, we check if there are any pending signals, and if so, we
		// execute their associated slots:
		qApp->processEvents();

		qDebug() << "QVWorker::iterate(): iteration" << numIterations;
	
		// Avoids "apparent hanging" (greedy ocupation of CPU by extremely fast
		// workers, such as paused ones). It is just 0.5 milliseconds, so
		// it should not be appreciable in any practical situation.
		usleep(500);

		switch (status)
			{
			case RunningOneStep:
				qDebug() << "QVWorker::iterate(): RunningOneStep";
				status = Paused;

			case Running:
				iterationTime.start();
				foreach(QList<QVPropertyContainer *> level, slavesByLevel) {
					foreach(QVPropertyContainer * slave, level) {
						QVWorker* worker;
						if((worker = dynamic_cast<QVWorker*>(slave)) != NULL) {
							worker->workerIterate();
						}
					}
				}
				curms = iterationTime.elapsed();
				if(minms > curms)
					usleep(1000*(minms-curms));
				/*if(numIterations!=1) // First iteration time is too noisy:
					acumms = (acumms*(numIterations-2) + curms) / (numIterations-1);
				std::cout << "-----> curms=" << curms << " acumms=" << acumms << "\n";*/
				break;

			case Stoped:
				// A stopped worker should never block anybody, but keeps linked to
				// other workers (otherwise, it would simply be deleted). So, it must
				// read its inputs an write its outputs always, even being stopped.
				readInputProperties();
				writeOutputProperties();
				usleep(100); // This avoids spurious CPU consuming when stopped.
				break;

	
			case Paused:
				qDebug() << "QVWorker::iterate(): Paused";
				usleep(100); // This avoids spurious CPU consuming when paused.
				break;

			case Finished:
				qDebug() << "QVWorker::iterate(): Finished";
				break;
			}

		if (maxIterations != -1 && numIterations >= maxIterations)
			finish();

		qDebug() << "QVWorker::iterate() <- return";
		}

	unlink();
	qDebug() << "QVWorker::run() <- return";
	}

void QVWorker::workerIterate()
	{
	if (statsEnabled) cpuStatControler->step();
	emit startIteration(); // if it is connect to a camera, it will be wait the camera grab()
	readInputProperties();
	timeFlag("System");
	if (status != Stoped) iterate();
	if (statsEnabled) setPropertyValue<QVStat>("cpu stats", cpuStatControler->value());
	writeOutputProperties();
	numIterations++;
	emit endIteration(getId(), getIteration());
	}

