/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvnumericplot.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef NUMERICPLOT_H
#define NUMERICPLOT_H

#include <qvgui/qvplot.h>

/*!
@class QVNumericPlot qvgui/qvnumericplot.h
@brief Widget for plotting integer and double property values from a worker, over the time.

This is a widget whichs plots output data values over a timeline:

@image html qvision_numericplot_window.png

For further info see section @ref QVNumericPlotWidget. 

@see QVHistogramPlot
@see QVCPUPlot

@ingroup qvgui
*/

class QVNumericPlot : public QVPlot
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVNumericPlot with a given string identifier, and iteration's parameters.
		/// Construct a new plot that shows int and double properties graphs. This plot shows decorations, it hasn't got a pie graph, don't brush the data
		/// and autoshow. It can show graphs of int and double linked properties of different workers.
		///
		/// @param name Identifier for the object to be created.
		/// @param time True if the iterations are based on clock time (in second hundredths), and false if they are based on worker's iterations.
		/// @param step Set the base measure to plot's iterations ratio.
		/// @param parent Identifier for the parent widget.
		QVNumericPlot(const QString name = QString (), bool time = true, int step = 10, QWidget *parent = 0);

		/// @brief Links properties from workers.
		///
		/// This method safely links properties from workers in a thread safe manner.
		/// If the iterations are based on worker's iterations (time == false), the worker must be synchronised with previous linked workers.
		/// Only can be linked int or double properties.
		///
		/// See also QVPropertyContainer::LinkType.
		/// @param worker The source QVWorker.
		/// @param propertyName Identifying QString for the source property.


		/// @brief Links properties from workers.
		///
		/// This method safely links all int and double output properties from a container with
		/// news properties with the same name, holders in a thread safe manner.
		///
		/// See also linkProperty(QVWorker &worker, const QString propertyName).
		/// @param worker The source QVWorker.


	protected:
		bool linkUnspecifiedInputProperty(QVPropertyContainer *sourceContainer, QString sourcePropName, LinkType linkType = AsynchronousLink);

	private:
		/// @brief Returns a list with a int or double property curve's names.
		/// Return a list with the name of the given property, that only contains one curve.
		///
		/// @param property The property name.
		QStringList getPropertyCurvNames(QString property) const;

		/// @brief Returns a list with a int or double property curve's value.
		/// Return a list with the value of the given property, that only contains one curve.
		///
		/// @param property The property name.
		QList<double> getPropertyCurvValues(QString property) const;

		/// @brief Returns a list with a int or double property curve's index.
		/// Return a list with the value 1 for the given property order, that only contains one curve, the first.
		///
		/// @param property The property name.
		QList<int> getPropertyCurvOrders(QString property) const;
	};

#endif

