/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcameraworker.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVCAMERAWORKER_H
#define QVCAMERAWORKER_H

#include <QVWorker>
#include <QVImage>

/*!
@class QVCameraWorker qvcameraworker.h qvio/qvcameraworker.h
@brief Virtual base class to create video reader "input worker blocks" for the QVision application architecture.

This class is pure virtual, and should not be used by normal QVision users (except when implementing their own video sources workers, such as, say, direct driver-based IEEE 1394 or V4L2 cameras; remember, though, that @ref QVMPlayerCameraWorker can read almost any kind of video input source (including live cameras, such as webcams, DV cameras, and so on). Any of its two derived classes @ref QVYUV4MPEG2CameraWorker and @ref QVMPlayerCameraWorker are the classes that normal users will use. See documentation on those classes for more details.

Implementors of new kind of QVCameraWorkers should only have to redefine this three pure virtual methods:

- virtual bool openCam(QString urlName, int &cols, int &rows, int &fps) = 0;
- virtual void closeCam() = 0;
- virtual bool grab(QVImage<uChar,1> &imgY, QVImage<uChar,1> &imgU, QVImage<uChar,1> &imgV) = 0;

Whose desired behaviour should be self-explanatory: the first one should open the video source specified by the string urlName, trying to open it with the suggested cols, rows and fps. The method must indicates a successful opening by returning true (and updating the corresponding cols, rows and fps output values to the really obtained ones, which could coincide or not with the suggested input values), or false, if there was any problem opening the source. The second must close the video source. Finally, the third must grab Y, U and V channel images in the corresponding variable parameters, with respective sizes cols x rows, cols/2 x rows/2 and cols/2 x rows/2 (with cols and rows as returned by the openCam method), returning true if grabbing was succesful, or false otherwise (for example, if the camera was closed).

All camera workers inheriting from this class will always have the following properties:

Input properties:

- <i>NoLoop</i> (bool): If the camera should be opened in no loop mode.
- <i>RealTime</i> (bool): If the camera should be opened in real time mode.
- <i>URL</i> (QString): URL of the video file to read.
- <i>Cols</i> (int): Number of suggested columns for the video source.
- <i>Rows</i> (int): Number of suggested rows for the video source.

<b>Note</b>.- RealTime property is also input (and as such can be initialized using the command line), but invisible to the GUI (because once the camera has been linked, it cannot be relinked changing its synchronism).

<b>Note</b>.- Note that you can open the camera in a different size of that specified in the real video source file (this class will automatically rescale output images if needed).

Output properties are the following:

- <i>Opened</i> (bool): Tells if the camera is correctly opened and working.
- <i>FPS</i> (int): FPS of the video.
- <i>Frames</i> (int): Number of read frames up to now -not necessarily equal to the number of processed frames, if in real time mode-.
- <i>ColsR</i> (int): Actual number of columns of the video -not necessarily equal to the number of requested columns, which will in any case be transparently rounded to the nearest upper even integer-.
- <i>RowsR</i> (int): Actual number of rows of the video -not necessarily equal to the number of requested rows, which will in any case be transparently rounded to the nearest upper even integer-..

<b>Note</b>.- Remember that you can open the camera in a different size of that specified in the real video source file (this class will automatically rescale output images if needed); thus, <i>ColsR</i> and <i>RowsR</i> will always contain the real size of the output images served by this worker.

Output image properties:

- <i>RGB image</i> (QVImage<uChar,3>): Last grabbed RGB image.
- <i>Y channel image</i> (QVImage<uChar,1>): Last grabbed Y channel image.
- <i>U channel image</i> (QVImage<uChar,1>): Last grabbed U channel image.
- <i>V channel image</i> (QVImage<uChar,1>): Last grabbed V channel image.

Output image properties can be linked automatically without the need of naming them explicitly, using the @ref QVCameraWorker::linkUnspecifiedOutputProperty() method.

@ingroup qvio
*/

class QVCameraWorker: public QVWorker
	{
	Q_OBJECT
	public:
		/// @brief Constructor for (virtual) QVCameraWorker class.
		///
		/// @param name the (optional) name given to the camera worker.
		QVCameraWorker(QString name = QString(""));

		/// @brief Destructor for (virtual) QVCameraWorker class.
		~QVCameraWorker();

		/// @brief Links a QVImage<uChar,3> or QVImage<uChar,1> from the camera to another worker.
		///
		/// This method will automatically link, using the adequate synchronism (Asynchronous for real time cameras
		/// and Synchronous for non real time cameras), a gray or an RGB image in the worker to the corresponding output
		/// image property of the camera. An error (return value=FALSE) will occur if dstProp is not of the adequate
		/// type (QVImage<uChar,3> or QVImage<uChar,1>).
		///
		/// @param dstCont Pointer to the destination worker.
		/// @param dstProp Name of the destination property.
		/// @return TRUE if success, false otherwise.
		bool linkUnspecifiedOutputProperty(QVPropertyContainer *dstCont, QString dstProp);

		/// @brief Links three QVImage<uChar,1> from the camera to the corresponding properties of another worker.
		///
		/// This method will automatically link, using the adequate synchronism (Asynchronous for real time cameras
		/// and Synchronous for non real time cameras), a set of three one channel images in the worker to the
		/// corresponding output image properties of the camera. An error (return value=FALSE) will occur if any of
		/// the dstProp are not of the adequate QVImage<uChar,1> type.
		///
		/// @param dstCont Pointer to the destination worker.
		/// @param dstProp1 Name of the destination property for image Y.
		/// @param dstProp2 Name of the destination property for image U.
		/// @param dstProp3 Name of the destination property for image V.
		/// @return TRUE if success, false otherwise.
		bool linkUnspecifiedOutputProperty(QVPropertyContainer *dstCont, QString dstProp1, QString dstProp2, QString dstProp3);

	protected:
		bool realTime, noLoop;

		/// @brief Pure virtual openCam method.
		///
		/// Reimplementations of this method should open the video source specified by the string urlName, 
		/// trying to open it with the suggested cols, rows and fps. The method must indicates a successful opening 
		/// by returning true (and updating the corresponding cols, rows and fps output values to the really obtained
		/// ones, which could coincide or not with the suggested input values), or false, if there was any problem
		/// opening the source.
		/// @param urlName QString denoting the video source.
		/// @param cols Suggested number of columns (as input). On return contains real number of columns (as output).
		/// @param rows Suggested number of rows (as input). On return contains real number of rows (as output).
		/// @param fps Suggested fps (as input). On return contains real fps (as output).
		/// @return TRUE if success, FALSE if fail.
		virtual bool openCam(QString urlName, int &cols, int &rows, int &fps) = 0;

		/// @brief Pure virtual closeCam method.
		///
		/// Reimplementations of this method should close the video source. 
		virtual void closeCam() = 0;

		/// @brief Pure virtual grab method.
		///
		/// Reimplementations of this method should grab Y, U and V channel images in the corresponding variable
		/// parameters, with respective sizes cols x rows, cols/2 x rows/2 and cols/2 x rows/2 (with cols and rows as
		/// returned by the @ref openCam() method), returning true if grabbing was succesful, or false otherwise 
		/// (for example, if the camera was closed).
		/// @param imgY On return, contains the grabbed Y image.
		/// @param imgU On return, contains the grabbed U image.
		/// @param imgV On return, contains the grabbed V image.
		virtual bool grab(QVImage<uChar,1> &imgY, QVImage<uChar,1> &imgU, QVImage<uChar,1> &imgV) = 0;

	private:
		bool newGrabbedFrame,flush_pending_images;
		QString urlName;
		QVImage<uChar,1> imageY, imageU, imageV;

		void tryOpeningCam();
		void resetCameraWorker();
		void iterate();
		void processTrigger(QString str);

	private slots:
		void grabFrame();
	};

#endif
