/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayercamera.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>

#include <QApplication>
#include <QDebug>
#include <QObject>
#include <QStringList>
#include <QRegExp>
#include <QMutex>

#include <qvipp.h>
#include <qvio.h>

#include <QVWorker>
#include <QVMPlayerCamera>

/***************************** QVMPlayerCamera Class **************************/
QVMPlayerCamera::QVMPlayerCamera(QString name):QVCamera(name), mplayerProxy(name+"peo")
	{
	qDebug() << "QVMPlayerCamera::QVMPlayerCamera()";
	if (qvApp == NULL)
		{
		QString str = "QVWorker::QVWorker(): the QVWorker cannot be created before the QVApplication instance. Aborting now.";
		std::cerr << qPrintable(str) << std::endl;
		exit(1);
		}

	connect(&mplayerProxy, SIGNAL(newGrab()), this, SIGNAL(newGrab()));
	connect(&mplayerProxy, SIGNAL(newRead()), this, SIGNAL(newRead()));
	connect(&mplayerProxy, SIGNAL(camClosed()), this, SIGNAL(camClosed()));
	connect(&mplayerProxy, SIGNAL(camOpened()), this, SIGNAL(camOpened()));
	connect(&mplayerProxy, SIGNAL(statusChange(QVCamera::TCameraStatus)), this, SIGNAL(statusChange(QVCamera::TCameraStatus)));

	rgbMode = false;

	// Properties (to allow reading from argc and argv):
	addProperty<int>("Rows", inputFlag, 0, "Rows to open the camera");
	addProperty<int>("Cols", inputFlag, 0, "Columns to open the camera");
	addProperty<bool>("RealTime", inputFlag, FALSE,"If the camera should be opened in real time mode");
	addProperty<bool>("Deinterlaced", inputFlag, FALSE,"If the camera should be opened in deinterlaced mode");
	addProperty<bool>("NoLoop", inputFlag, FALSE,"If the camera should be opened in no loop mode");
	addProperty<QString>("URL", inputFlag, QString(""),"URL of the video source");

	// QVImage's
	addProperty< QVImage<uChar,3> >("RGB image", outputFlag, QVImage<uChar,3>(), "Last grabbed RGB image");
	addProperty< QVImage<uChar,1> >("Y channel image",outputFlag, QVImage<uChar,1>(), "Last grabbed Y channel image");
	addProperty< QVImage<uChar,1> >("U channel image",outputFlag, QVImage<uChar,1>(), "Last grabbed U channel image");
	addProperty< QVImage<uChar,1> >("V channel image",outputFlag, QVImage<uChar,1>(), "Last grabbed V channel image");

	qDebug() << "QVMPlayerCamera::QVMPlayerCamera() <- return";
	}

QVMPlayerCamera::~QVMPlayerCamera()
	{
	qDebug() << "QVMPlayerCamera::~QVMPlayerCamera()";
	if (status != Closed) closeCam();
	qDebug() << "QVMPlayerCamera::~QVMPlayerCamera() <- deleting image";
	qDebug() << "QVMPlayerCamera::~QVMPlayerCamera() <- return";
	}

// (Implicit) default open method for a camera (uses introspection to read 
// parameters, allowing initialization from argc and argv):
bool QVMPlayerCamera::openCam()
	{
	std::cout << "QVMPlayerCamera::openCam()." << std::endl;
	QVMPlayerProxy::OpenOptions optsProp = QVMPlayerProxy::Default;

	readInputProperties();

	if (getPropertyValue<bool>("RealTime"))		optsProp |= QVMPlayerProxy::RealTime;
	if (getPropertyValue<bool>("Deinterlaced"))	optsProp |= QVMPlayerProxy::Deinterlaced;
	if (getPropertyValue<bool>("NoLoop"))		optsProp |= QVMPlayerProxy::NoLoop;
	if (rgbMode)					optsProp |= QVMPlayerProxy::RGBMEncoder;

	return mplayerProxy.openCam(getPropertyValue<QString>("URL"), optsProp, getPropertyValue<int>("Rows"), getPropertyValue<int>("Cols"));
	}

// Close camera method:
void QVMPlayerCamera::closeCam()
	{
	mplayerProxy.closeCam();
	}

bool QVMPlayerCamera::grab()
	{
	qDebug() << "QVMPlayerCamera::grab()";

	QVImage<uChar> grabbedYImage, grabbedUImage, grabbedVImage;

	if (!mplayerProxy.grab(grabbedYImage, grabbedUImage, grabbedVImage))
		return false;

	setPropertyValue< QVImage<uChar,1> >("Y channel image", grabbedYImage);
	setPropertyValue< QVImage<uChar,1> >("U channel image", grabbedUImage);
	setPropertyValue< QVImage<uChar,1> >("V channel image", grabbedVImage);

	if (rgbMode)
		{
		QVImage<uChar,3> grabbedRGBImage(grabbedYImage.getCols(),grabbedYImage.getRows());
		YUV420ToRGB(grabbedYImage, grabbedUImage, grabbedVImage, grabbedRGBImage);
		setPropertyValue< QVImage<uChar,3> >("RGB image", grabbedRGBImage);
		}

	writeOutputProperties();

	qDebug() << "QVMPlayerCamera::grab() <~ return";

	return true;
	}

//////////////////////////
bool QVMPlayerCamera::linkProperty(QString, QVPropertyContainer *destinyContainer, QString destinyPropertyName, LinkType linkType)
	{
		std::cerr << "Warning: QVMPlayerCamera::linkProperty(): undefined specific link, the sourcePropertyName will be ignored" << std::endl;
		return linkProperty(destinyContainer, destinyPropertyName, linkType);
	}

bool QVMPlayerCamera::linkProperty(QString sourcePropertyName, QVPropertyContainer &destinyContainer, QString destinyPropertyName, LinkType linkType)
	{
		return linkProperty(sourcePropertyName, &destinyContainer, destinyPropertyName, linkType);
	}

bool QVMPlayerCamera::linkProperty(QVPropertyContainer *destinyContainer, QString destinyPropertyName, LinkType linkType)
	{
	if (linkType == SynchronousLink) {
		std::cerr << "ERROR: QVMPlayerCamera::linkProperty(): the linkType must be AsynchronousLink, the link will not be done" << std::endl;
		return FALSE;
	}

	QVWorker* worker;
	if((worker = dynamic_cast<QVWorker*>(destinyContainer)) != NULL)
		{
		if(worker->isType<QVImage<uChar,1> >(destinyPropertyName))
			{
			connect(worker, SIGNAL(startIteration()), this, SLOT(grab()));
			return QVPropertyContainer::linkProperty("Y channel image", worker, destinyPropertyName, AsynchronousLink);
			}
		else if(worker->isType<QVImage<uChar,3> >(destinyPropertyName))
			{
			rgbMode = true;
			connect(worker, SIGNAL(startIteration()), this, SLOT(grab()));
			return QVPropertyContainer::linkProperty("RGB image", worker, destinyPropertyName, AsynchronousLink);
			}
		else
			{
			std::cerr << "Warning: QVMPlayerCamera::linkProperty(): unsupported property type: property " << qPrintable(destinyPropertyName)
				<< " in holder " << qPrintable(worker->getName()) << "is not a QVImage<uChar,1> or QVImage<uChar,3>" << std::endl;
			return FALSE; //Should never get here;
			}
		}
	else
		{
		std::cerr << "ERROR: QVMPlayerCamera::linkProperty(): the destination container must be a QVWorker, the link will not be done" << std::endl;
		return FALSE;
		}
	}

bool QVMPlayerCamera::linkProperty(QVPropertyContainer &destinyContainer, QString destinyPropertyName, LinkType linkType)
	{
		return linkProperty(&destinyContainer, destinyPropertyName, linkType);
	}

bool QVMPlayerCamera::linkProperty(QString, QVPropertyContainer *, LinkType)
	{
	std::cerr << "ERROR: QVMPlayerCamera::linkProperty(): the destination property name can't be omitted, the link will not be done" << std::endl;
	return FALSE;
	}

bool QVMPlayerCamera::linkProperty(QString sourcePropertyName, QVPropertyContainer &destinyContainer, LinkType linkType)
	{
		return linkProperty(sourcePropertyName, &destinyContainer, linkType);
	}

void QVMPlayerCamera::linkProperty(QVPropertyContainer *, LinkType)
	{
	std::cerr << "ERROR: QVMPlayerCamera::linkProperty(): unsopported global link, the link will not be done" << std::endl;
	}

void QVMPlayerCamera::linkProperty(QVPropertyContainer &container, LinkType linkType)
	{
		linkProperty(&container, linkType);
	}

bool QVMPlayerCamera::unlinkProperty(QString origName, QVPropertyContainer *destCont, QString destName)
	{
	QVWorker* worker;
	if((worker = dynamic_cast<QVWorker*>(destCont)) != NULL)
		{
		if (origName == "RGB image")
			{
			disconnect(worker, SIGNAL(startIteration()), this, SLOT(grab()));
			return QVPropertyContainer::unlinkProperty(origName, worker, destName);
			}
		else if (origName == "Y channel image")
			{
			disconnect(worker, SIGNAL(startIteration()), this, SLOT(grab()));
			return QVPropertyContainer::unlinkProperty(origName, worker, destName);
			}
		else
			{
			std::cerr << "ERROR: QVMPlayerCamera::unlinkProperty(): invalid source property, the unlink will not be done" << std::endl;
			return FALSE;
			}
		}
	else
		{
		std::cerr << "ERROR: QVMPlayerCamera::unlinkProperty(): the destination container must be a QVWorker, the unlink will not be done" << std::endl;
		return FALSE;
		}
	}

bool QVMPlayerCamera::unlinkProperty(QString origName, QVPropertyContainer &destCont, QString destName)
	{
		return unlinkProperty(origName, &destCont, destName);
	}

void QVMPlayerCamera::unlink()
	{
		///@todo disconnect(); //deberia desconectarse, pero ¿podría estar conectada a varios workers?
		QVPropertyContainer::unlink();
	}

bool QVMPlayerCamera::link(QVWorker* worker, const QString imageNameY, const QString imageNameU, const QString imageNameV)
	{
	if(!worker->isType< QVImage<uChar,1> >(imageNameY))
		{
		std::cerr << "Warning: QVMPlayerCamera::link(): image " << qPrintable(imageNameY) << "is not of type QVImage<uChar,1> in holder "
			<< qPrintable(worker->getName()) << std::endl;
		return FALSE;
		}

	if(!worker->isType< QVImage<uChar,1> >(imageNameU))
		{
		std::cerr << "Warning: QVMPlayerCamera::link(): image " << qPrintable(imageNameU) << "is not of type QVImage<uChar,1> in holder "
			<< qPrintable(worker->getName()) << std::endl;
		return FALSE;
		}

	if(!worker->isType< QVImage<uChar,1> >(imageNameV))
		{
		std::cerr << "Warning: QVMPlayerCamera::link(): image " << qPrintable(imageNameV) << "is not of type QVImage<uChar,1> in holder "
			<< qPrintable(worker->getName()) << std::endl;
		return FALSE;
		}

	connect(worker, SIGNAL(startIteration()), this, SLOT(grab()));
	return	this->linkProperty("Y channel image", worker, imageNameY, AsynchronousLink) &&
		this->linkProperty("U channel image", worker, imageNameU, AsynchronousLink) &&
		this->linkProperty("V channel image", worker, imageNameV, AsynchronousLink);
	}

bool QVMPlayerCamera::link(QVWorker &worker, const QString imageNameY, const QString imageNameU, const QString imageNameV)
	{
		return link(&worker, imageNameY, imageNameU, imageNameV);
	}

bool QVMPlayerCamera::unlink(QVWorker* worker, const QString imageName)
	{
		if (rgbMode) {
			return unlinkProperty("RGB image", worker, imageName);
		}
		else {
			return unlinkProperty("Y channel image", worker, imageName);
		}
	}

bool QVMPlayerCamera::unlink(QVWorker &worker, const QString imageName)
	{
		return unlink(&worker, imageName);
	}

