/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmplayerproxy.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVMPLAYERPROXY_H
#define QVMPLAYERPROXY_H

#include <QProcess>
#include <QString>
#include <QUrl>
#include <QThread>
#include <QMutex>
#include <QWaitCondition>
#include <QFile>
#include <QTimer>
#include <QQueue>

#include <QVCamera>
#include <qvutils/qnamedpipe.h>

#ifndef DOXYGEN_IGNORE_THIS
// Auxiliary QVCheckOKMPlayerCamera Class.
// This is an internal convenience thread used only to unblock the open function
// when it is waiting for mplayer to read from the fifo, in the case when
// mplayer aborted when starting (due to bad URL, or wrong file format, for
// example).

class QVCheckOKMplayerCamera: public QThread
{
	Q_OBJECT
	public:
		QVCheckOKMplayerCamera(QFile & fifo_file,int max_time_ms_to_wait_for_open);
		void run();
	private slots:
		void writeErrorInFifo();
	private:
		QFile & _fifo_file;
		int _max_time_ms_to_wait_for_open;
};

// Auxiliary QVCameraThread Class.
// This is an internal convenience thread used in real time cameras to keep
// updating the frames from mplayer independently of the user grabs.

class QVCameraThread: public QThread
	{
	Q_OBJECT
	public:
		QVCameraThread(QObject *object,char *slot);
		~QVCameraThread();
		void run();
	private:
		QObject *_object;
		char *_slot;
	};

// Auxiliary QVMPlayerIOProcessor class.
// Auxiliary class to communicate through stdin and stdout with mplayer, and
// read data such as the number of rows and cols of the video, the fps, and also
// to send commands such as pausing, ask for current position, and so on.

class QVMPlayerIOProcessor: public QObject
	{
	Q_OBJECT
	public:
		QVMPlayerIOProcessor(QProcess *mplayer);
		~QVMPlayerIOProcessor();
		void queueCommandToMPlayer(const QString &, bool ifEmpty = false);
		double fps, speed, time_length, time_pos;
		uInt cols, rows;
	public slots:
		int interpretMPlayerOutput();
	private slots:
		void sendCommandToMPlayer();
	private:
		QQueue<QString> command_queue;
		QMutex command_queue_mutex;
		QProcess *mplayer;
	};

// Auxiliary QVMPlayerFrameGrabber Class.
class QVMPlayerFrameGrabber: public QObject
	{
	Q_OBJECT

	public:
		QVMPlayerFrameGrabber(	QVMPlayerIOProcessor *, QString pipeName, bool realTime, const uInt waitMilisecs);
		~QVMPlayerFrameGrabber();

		void getQVImageGray(QVImage<uChar> &imgGray) const;
		void getQVImageYUV(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV) const;

		void updateFrameBuffer();

		bool isFinished()	const	{ return finished; }
		bool framesRead()	const	{ return frames_read; }

	public slots:
		void updateSlot();

	signals:
		void newReadFrameGrabber();
		// This signal is to defer the call to sendCommandToMplayer, and avoid
		// the "QSocketNotifier:: socket notifier cannot be enabled from another
		// thread" warning:
		void sendCommandSignal();

	private:
		QVCameraThread *qvcamerathread;
		QFile fifoInput;	
		QVMPlayerIOProcessor *mplayerIOProcessor;

		QMutex mutex;
		QWaitCondition condition;
		
		uInt frames_read;
		bool finished, realTime, inited;
		//bool debug_in_memcpy, debug_in_updateSlot;

		QVImage<uChar> imgY, imgU, imgV, img_auxY, img_auxU,  img_auxV;
	};
#endif

// This is for doxygen get the detailed description for QVMPlayerProxy correctly.
//class QVWorker;

/*!
@class QVMPlayerProxy qvmplayerproxy.h qvio/qvmplayerproxy.h
@brief Class to read from many video formats and devices, based on the MPlayer application.

This class is based on the <a href="http://www.mplayerhq.hu/">MPlayer</a> application. It can read videos from a wide variety of media formats, webcams and digital cameras.

Internally, when used to read from a video source compatible with MPlayer, the @ref QVMPlayerProxy object will launch the MPlayer application to open and read from a video source. The MPlayer will be configured to redirect its output to a pipeline, from which the @ref QVMPlayerProxy class will read the input frames.
deo source.

\section QVMPlayerURLFormat Video source identifier URL formats
Each video source is identified by a URL string, even for video files, webcams, remote videos, etc... this class opens a video source, and reads frames from it, until a stop condition arises or the video finishes.

\subsection URLSimplifiedFormat Simplified camera URL format.
Format: just the path to a file or files. Type of formats
 - For video files: 
	\verbatim path-to-file/file-name.{avi, mpg, dv, ...} \endverbatim
 - For devices (webcams, digital cams, tv inputs):
	\verbatim /dev/devicename \endverbatim
 - For image lists:
	\verbatim path-to-directory/filemask.{jpg, png, ...} \endverbatim

\subsection ExtendedFormat Extended camera URL formats.
With previous format there can be incompleteness, ambiguous video input, and other problems. More specific URL formats follows:
The format of this string depends of the type of the video source.
Notation: [] means optional, {} means alternative:
	- Local video file compatible with mplayer:
		\verbatim file://[path-to-file/]file-name.{avi, mpg, dv, ...} \endverbatim
		Path can be relative, or absolute. Absolute paths begin with an extra slash. Examples:
		\verbatim file:///home/vision/penguin.dv \endverbatim
		\verbatim file://../../home/vision/penguin.dv \endverbatim
	- Remote video file (via http, rtsp, ftp):
		\verbatim {http,rtsp,ftp}://domain-name/path-to-file/file.{avi, mpg, dv, ...} \endverbatim
		Examples:
		\verbatim rtsp://ficheros.7rm.es/7RM/7rm06-01-071430.rm \endverbatim
		\verbatim http://perception.inf.um.es/public_data/videos/misc/penguin.dv \endverbatim
		\verbatim ftp://user:passwd@server.dom/data/video.avi \endverbatim

	- DVD | VCD video track (see mplayer man page for correct URL 
		formation for a specific title, track, device, etc.).		
		\verbatim {dvd,vcd}://{video track}\endverbatim
		Examples:
		\verbatim dvd://2 \endverbatim
		\verbatim vcd://1 \endverbatim

	- Sequence of multiple JPEG or PNG files:
		\verbatim mf://path-to/filemask.{jpg, png, ...} \endverbatim
		Examples:
		\verbatim mf:///home/user/images/img*.jpg \endverbatim
		\verbatim mf:///./images/img*.png \endverbatim

	- Video 4 linux, video 4 linux 2 camera, and analog inputs:
		\verbatim {v4l,v4l2,analog}://{device file}?[options] \endverbatim
		Examples:
		\verbatim v4l:///dev/video0 \endverbatim
		\verbatim analog:///dev/video0 \endverbatim
		\b Note: Optional specifiers H and W are used by mplayer to open
		the camera in a given source size. Note that this is not the
		same as opening the camera in a default size and then scaling
 		the output image to a given size using the \c r and \c c
		parameters (see overloaded function).
		Check manual page for mplayer, for a complete list of options, at parameter '-tv' option
		specifications.

\subsection CameraParameters Camera parameters
Both extended and simplified formats can include special parameters for tuning device input options. Some of these parameters are:
	- <b>width=&lt;value&gt;</b>	<br>sets the width of the image frames to be read from the video device.
	- <b>height=&lt;value&gt;</b>	<br>sets the height of the image frames to be read from the video device.
	- <b>fps=&lt;value&gt;</b>	<br>sets the number of frames per second to be read from the video device.
	- <b>[brightness|contrast|hue|saturation]=[-100, 100]</b>
					<br>sets the brightness, contrast, hue or saturation for the device.
	- <b>channel=&lt;value&gt;</b>	<br>specifies the tv channel to read images from.
You can get a complete list of these options at the MPlayer documentation, about the user line input parameter <i>-tv</i>.

Setting the with and height for the images read from the input device using parameters <i>width</i> and <i>height</i> differs from doing so with options <i>--Cols</i> and <i>--Rows</i> of the input command line. The former can configure the device to capture images using a higher or lesser resolution, while the latter scale the image as MPlayer reads them from the input device.

These options work for video 4 Linux, video 4 Linux 2 camera, and analog inputs. They can be included appending a <i>?</i> sign after the URL for the device, and separated with <i>&</i> signs. Some example usages of these URL options follow:
	- URL to open a V4L2 device specifying 25 frames per second, image size of 640x480, and minimum brightness:
		\verbatim v4l2:///dev/video0?height=480&width=640&brightness=-100&fps=25 \endverbatim
	- The same device and configuration using a simplified URL:
		\verbatim /dev/video0?height=480&width=640&brightness=-100&fps=25 \endverbatim
	- URL to open an tv device at a 320x240 resolution, on channel 42
		\verbatim analog:///dev/tv?height=240&width=320&channel=42 \endverbatim

@see QVMPlayerCamera
@ingroup qvio
*/

class QVMPlayerProxy : public QObject
	{
	Q_OBJECT
	public:
		/// @brief Constructs a QVMPlayerProxy, which is initially in "Closed" status.
		QVMPlayerProxy(QString name = QString());
		/// @brief Destroys a QVMPlayerProxy.
		~QVMPlayerProxy();

		// Camera opening options as flags. Combine using | operator to create
		// an OpenOptions flags (observe plural) object:
		/// @brief Open options for a mplayer camera. Combine them using OR (|).
		enum OpenOption {
			/// Default options: don't include any of the other options.
			Default = 0x0,
			/// Camera runs in an independent, real time thread.
			RealTime = 0x1,
			/// Make mplayer deinterlace the image (odd lines).
			Deinterlaced = 0x2,
			/// Continue from the beginning of the video when it ends.
			NoLoop = 0x4,
			/// Camera will produce interleave RGB images (instead of YUV).
			RGBMEncoder = 0x8
		};

		Q_DECLARE_FLAGS(OpenOptions,OpenOption);

		/// @brief Opens a mplayer source given by its URL, and scales output to a requested size.
		///
		///	 (Overloaded function, provided by convenience).
		///	 @param url Camera identifying URL string. See @ref QVMPlayerURLFormat for URL formats.
		///	 @param r Number of requested output rows.
		///	 @param c Number of requested output cols.
		///	 @param opts Camera opening options (see QVMPlayerProxy::openCam(const QString&,OpenOptions))
		/// @return True if the camera was opened successfully, false otherwise.
		bool openCam(const QString & urlstring, OpenOptions opts = Default, unsigned int r = 0, unsigned int c = 0);

		/// @brief Grabs a frame in RGB format.
		///
		/// If the camera is not opened with RGBMencoderMode option, this function will set parameter QVImage to null image.
		/// @param image QVImage that will store grabbed frame.
		/// @returns TRUE if success, FALSE if fail.
		bool grab(QVImage<uChar,3> &image);

		/// @brief Grabs a gray scale frame.
		///
		/// If the camera is opened with RGBMencoderMode option, this function will set param QVImage to null image.
		/// @param image QVImage that will store grabbed frame.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar,1> &image);

		/// @brief Grabs a frame in YUV format.
		///
		/// This function grabs the image in YUV format, storing each channel Y, U and V, in a separate QVImage of
		/// one channel. YUV format would be 422, thus QVImage corresponding to plane Y will be four times bigger than
		/// those corresponding to planes U and V.
		///
		/// If the camera is opened with RGBMencoderMode option, this function will set param QVImage's to null images.
		/// @param imgY QVImage that will store grabbed frame Y channel.
		/// @param imgU QVImage that will store grabbed frame U channel.
		/// @param imgV QVImage that will store grabbed frame V channel.
		/// @returns True if success, false otherwise.
		bool grab(QVImage<uChar> &imgY, QVImage<uChar> &imgU, QVImage<uChar> &imgV);

		/// @brief Returns the current camera name (without path).
		/// @returns Current camera name.
		/// @deprecated Use getName() instead
		const QString getUrlBase() const		{ return path.split("/").last(); }

		/// @brief Returns current camera options.
		/// @returns Current camera options.
		OpenOptions getOptions() const			{ return open_options; };

		/// @brief Returns the total number of updated frames.
		/// @returns Current total number of updated frames.
		unsigned int getFramesUpdated() const		{ return mplayerFrameGrabber->framesRead(); }

		/// @brief Returns the number of rows of the buffer.
		/// @returns Number of rows.
		unsigned int getRows() const			{ return mplayerIOProcessor->rows; };

		/// @brief Returns the number of cols of the buffer.
		/// @returns Number of cols.
		unsigned int getCols() const			{ return mplayerIOProcessor->cols; };

		/// @brief Returns the camera speed in frames per second.
		/// @returns Camera speed in frames per second.
		double getFPS() const				{ return mplayerIOProcessor->fps; };

		/// @brief Returns the video file length in seconds.
		/// @returns Video length in seconds.
		double getTimeLength() const			{ return mplayerIOProcessor->time_length; };

		/// @brief Returns the current video position in seconds.
		/// @returns Current video position in seconds.
		double getTimePos() const			{ return mplayerIOProcessor->time_pos; };

		/// @brief Returns the current video speed.
		/// @returns Current video speed (as a multiplier of the FPS).
		double getSpeed() const				{ return mplayerIOProcessor->speed; };

		/// @brief Returns number of frames grabbed.
		/// @returns Number of successful frames returned by grab's
		unsigned int getFramesGrabbed() const		{ return frames_grabbed; };

		/// @brief Returns number of frames read.
		/// @returns Number of frames actually read by a camera.
		unsigned int getFramesRead() const		{ return mplayerFrameGrabber->framesRead(); };

		/// @brief Tells if the camera is a live camera.
		/// @returns True if camera is a live camera.
		bool isLiveCamera() const			{ return live_camera; };

		QVCamera::TCameraStatus getCameraStatus() const { return status; };
		bool isClosed() const { return status == QVCamera::Closed; };
		bool isRunning() const { return status == QVCamera::Running; };
		bool isRunningOneStep() const { return status == QVCamera::RunningOneStep; };
		bool isPaused() const { return status == QVCamera::Paused; };

	public slots:

		/// @brief Pauses the camera.
		void pauseCam()						{ pauseCamSlot(); }

		/// @brief Unpauses the camera.
		void unpauseCam()					{ unpauseCamSlot(); }

		/// @brief Go to the next frame.
		void nextFrameCam()					{ nextFrameCamSlot(); }

		/// @brief Set the camera speed.
		/// @param d Camera speed (as a multiplier of the FPS)
		void setSpeedCam(double d)				{ setSpeedCam(d); }

		/// @brief Moves the camera file to a given position.
		/// @param type Type of seek: Current=0,Percentage=1,Absolute=2.
		/// @param pos 
		///	- If in \c Current mode: Relative displacement with respect to the current position in seconds.
		///	- If in \c Percentage mode: Percentage of total video length.
		///	- If in \c Absolute mode: Absolute position in seconds.
		void seekCam(QVCamera::TSeekType type, double pos)	{ seekCamSlot(type, pos); }

		/// @brief Closes the camera.
		void closeCam();

		void pauseCamSlot();
		void unpauseCamSlot();
		void nextFrameCamSlot();
		void setSpeedCamSlot(double d);
		void seekCamSlot(QVCamera::TSeekType type, double pos);


	signals:
		// This signal is to defer the call to sendCommandToMplayer, and avoid
		// the "QSocketNotifier:: socket notifier cannot be enabled from another
		// thread" warning:
		void sendCommandSignal();
		void newGrab();
		void newRead();
		void camClosed();
		void camOpened();
		void statusChange(QVCamera::TCameraStatus);

	private:
		// this boolean value becomes true if any RGB image is linked to the camera, from a worker.
		//bool rgbMode;
		QVCamera::TCameraStatus status;

		OpenOptions open_options;
		QVMPlayerFrameGrabber * mplayerFrameGrabber;
		QVMPlayerIOProcessor * mplayerIOProcessor;

		QNamedPipe *namedPipe;			// The named fifo;
		QProcess mplayer;			// mplayer and mencoder processes;

		void initMPlayerArgs(QString urlString, unsigned int rows, unsigned int cols);
		bool performGrab();

		QStringList mplayer_args;		// MPlayer arguments
		QString path, schema;			// path and schema extracted from the URL
		
		bool live_camera;
		int frames_grabbed;

		void setStatus(QVCamera::TCameraStatus status) 
			{ this->status = status; emit statusChange(status); }
	};

// Camera opening flags | operator:
Q_DECLARE_OPERATORS_FOR_FLAGS(QVMPlayerProxy::OpenOptions)

#endif




